const MockDate = require('mockdate')
const util = require('../util')

describe('Sandbox Utilities', () => {
  describe('methods.getFiscalValue()', () => {
    it("null when key doesn't exist", () => {})
    it('correctly get value for annually', () => {
      const source = {
        test: {
          _type: 'annually',
          _values: [
            {
              fk: '2016',
              ck: '2017',
              v: 10
            }
          ]
        }
      }
      const fiscalKey = '2016'
      const result = util.getFiscalValue(source, 'test', fiscalKey)
      expect(result).toEqual(10)
    })
    it('correctly get value for quarterly', () => {
      const source = {
        test: {
          _type: 'quarterly',
          _values: [
            {
              fk: '2015Q4',
              ck: '2015Q3',
              v: 10
            }
          ]
        }
      }
      const fiscalKey = '2015Q4'
      const result = util.getFiscalValue(source, 'test', fiscalKey)
      expect(result).toEqual(10)
    })
    it('correctly get value for monthly', () => {
      const source = {
        test: {
          _type: 'monthly',
          _values: [
            {
              fk: '2016-12',
              ck: '2017-1',
              v: 10
            }
          ]
        }
      }
      const fiscalKey = '2016-12'
      const result = util.getFiscalValue(source, 'test', fiscalKey)
      expect(result).toEqual(10)
    })
  })
  describe('methods.getCalendarValue()', () => {
    it('correctly get value for annually', () => {
      const source = {
        test: {
          _type: 'annually',
          _values: [
            {
              fk: '2016',
              seen: '2017',
              v: 11
            }
          ]
        }
      }
      const calendarKey = '2017'
      const result = util.getCalendarValue(source, 'test', calendarKey)
      expect(result).toEqual(11)
    })
    it('correctly get value for quarterly', () => {
      const source = {
        test: {
          _type: 'quarterly',
          _values: [
            {
              fk: '2016Q4',
              seen: new Date(2012, 2, 31),
              v: 11
            }
          ]
        }
      }
      const calendarKey = '2017Q1'
      const result = util.getCalendarValue(source, 'test', calendarKey)
      expect(result).toEqual(11)
    })
    it('correctly get value for monthly', () => {
      const source = {
        test: {
          _type: 'monthly',
          _values: [
            {
              fk: '2016-12',
              seen: new Date(2017, 0, 1),
              v: 11
            }
          ]
        }
      }
      const calendarKey = '2017-1'
      const result = util.getCalendarValue(source, 'test', calendarKey)
      expect(result).toEqual(11)
    })
  })

  describe('util.dateToScopeKey(date, scopeKey)', () => {
    describe('date is end period', () => {
      it('converts date to daily key correctly', () => {
        const expected = '2015-3-15'
        const date = new Date(Date.UTC(2015, 2, 15, 0, 0, 0))
        const actual = util.dateToScopeKey(date, 'daily')
        expect(expected).toEqual(actual)
      })
      it('converts date to annually key correctly', () => {
        const expected = '2015'
        const date = new Date(Date.UTC(2015, 11, 31, 0, 0, 0))
        const actual = util.dateToScopeKey(date, 'annually')
        expect(expected).toEqual(actual)
      })
      it('converts date to quarterly key correctly', () => {
        const expected = '2015Q3'
        const date = new Date(Date.UTC(2015, 8, 30, 0, 0, 0))
        const actual = util.dateToScopeKey(date, 'quarterly')
        expect(expected).toEqual(actual)
      })
      it('converts date to monthly key correctly', () => {
        const expected = '2015-9'
        const date = new Date(Date.UTC(2015, 8, 30, 0, 0, 0))
        const actual = util.dateToScopeKey(date, 'monthly')
        expect(expected).toEqual(actual)
      })
      it('converts quarter to year correctly', () => {
        const expected = '2016'
        const date = new Date(Date.UTC(2016, 3, 31, 0, 0, 0))
        const actual = util.dateToScopeKey(date, 'annually')
        expect(expected).toEqual(actual)
      })
    })
    describe('date is NOT end period', () => {
      it('converts date to daily key correctly', () => {
        const expected = '2015-3-15'
        const date = new Date(Date.UTC(2015, 2, 15, 0, 0, 0))
        const actual = util.dateToScopeKey(date, 'daily')
        expect(expected).toEqual(actual)
      })
      it('converts date to annually key correctly', () => {
        const expected = '2015'
        const date = new Date(Date.UTC(2015, 11, 27, 0, 0, 0))
        const actual = util.dateToScopeKey(date, 'annually')
        expect(expected).toEqual(actual)
      })
      it('converts date to quarterly key correctly', () => {
        const expected = '2015Q3'
        const date = new Date(Date.UTC(2015, 8, 27, 0, 0, 0))
        const actual = util.dateToScopeKey(date, 'quarterly')
        expect(expected).toEqual(actual)
      })
      it('converts date to monthly key correctly', () => {
        const expected = '2015-9'
        const date = new Date(Date.UTC(2015, 8, 27, 0, 0, 0))
        const actual = util.dateToScopeKey(date, 'monthly')
        expect(expected).toEqual(actual)
      })
    })
  })

  describe('scopeKeyToDate()', () => {
    it('converts daily to that UTC date', () => {
      const testKey = '2015-03-15'
      const actual = util.scopeKeyToDate(testKey)
      expect(actual.getUTCDate()).toEqual(15)
      expect(actual.getUTCMonth()).toEqual(2)
      expect(actual.getUTCFullYear()).toEqual(2015)
      expect(actual.getUTCHours()).toEqual(0)
      expect(actual.getUTCMinutes()).toEqual(0)
      expect(actual.getUTCMilliseconds()).toEqual(0)
    })
    it('converts annually to last day of year', () => {
      const testKey = '2015'
      const actual = util.scopeKeyToDate(testKey)
      expect(actual.getUTCDate()).toEqual(31)
      expect(actual.getUTCMonth()).toEqual(11)
      expect(actual.getUTCFullYear()).toEqual(2015)
      expect(actual.getUTCHours()).toEqual(0)
      expect(actual.getUTCMinutes()).toEqual(0)
      expect(actual.getUTCMilliseconds()).toEqual(0)
    })
    it('converts quarterly to last day of quarter', () => {
      const testKey = '2015Q3'
      const actual = util.scopeKeyToDate(testKey)
      expect(actual.getUTCDate()).toEqual(30)
      expect(actual.getUTCMonth()).toEqual(8)
      expect(actual.getUTCFullYear()).toEqual(2015)
      expect(actual.getUTCHours()).toEqual(0)
      expect(actual.getUTCMinutes()).toEqual(0)
      expect(actual.getUTCMilliseconds()).toEqual(0)
    })
    it('converts monthly to last day of monthy', () => {
      const testKey = '2015-5'
      const actual = util.scopeKeyToDate(testKey)
      expect(actual.getUTCDate()).toEqual(31)
      expect(actual.getUTCMonth()).toEqual(4)
      expect(actual.getUTCFullYear()).toEqual(2015)
      expect(actual.getUTCHours()).toEqual(0)
      expect(actual.getUTCMinutes()).toEqual(0)
      expect(actual.getUTCMilliseconds()).toEqual(0)
    })
  })

  // describe.skip('difference()', () => {
  //   it('difference larger value with smaller value correctly', () => {
  //     const result = util.difference('2015Q4', '2013Q2')
  //     expect(result).toEqual(10)
  //   })
  //   it('difference smaller value with larger value correctly', () => {
  //     const result = util.difference('2010Q1', '2013Q2')
  //     expect(result).toEqual(-13)
  //   })
  //   it('difference equal value correctly', () => {
  //     const result = util.difference('2015Q4', '2015Q4')
  //     expect(result).toEqual(0)
  //   })
  // })

  describe('subtractScopeKey()', () => {
    it('subtract defaults to 1', () => {
      const result = util.subtractScopeKey('2014Q1')
      expect(result).toEqual('2013Q4')
    })
    it('subtract -1', () => {
      const result = util.subtractScopeKey('2014Q1', -1)
      expect(result).toEqual('2014Q2')
    })
    it('subtract 1', () => {
      const result = util.subtractScopeKey('2014Q1', 1)
      expect(result).toEqual('2013Q4')
    })
    it('subtract 0', () => {
      const result = util.subtractScopeKey('2014Q1', 0)
      expect(result).toEqual('2014Q1')
    })
  })

  describe('compareKey()', () => {
    describe('scope quarterly', () => {
      it('compare larger value with smaller value correctly', () => {
        const result = util.compareKey('2015Q4', '2013Q2')
        expect(result).toEqual(1)
      })
      it('compare smaller value with larger value correctly', () => {
        const result = util.compareKey('2010Q1', '2013Q2')
        expect(result).toEqual(-1)
      })
      it('compare equal value correctly', () => {
        const result = util.compareKey('2015Q4', '2015Q4')
        expect(result).toEqual(0)
      })
    })
  })

  describe('keysInRange()', () => {
    test('quarterly keys', () => {
      const result = util.keysInRange('2001Q3', '2002Q2')
      expect(result).toEqual(['2001Q3', '2001Q4', '2002Q1', '2002Q2'])
    })

    test('annually keys', () => {
      const result = util.keysInRange('2000', '2002')
      expect(result).toEqual(['2000', '2001', '2002'])
    })

    test('monthly keys', () => {
      const result = util.keysInRange('2016-12', '2017-6')
      expect(result).toEqual(['2016-12', '2017-01', '2017-02', '2017-03', '2017-04', '2017-05', '2017-06'])
    })

    test('no filling needed', () => {
      const result = util.keysInRange('2000', '2001')
      expect(result).toEqual(['2000', '2001'])
    })

    test('end and start is equal', () => {
      const result = util.keysInRange('2000', '2000')
      expect(result).toEqual(['2000'])
    })

    test('start greater than end key', () => {
      const result = util.keysInRange('2001', '2000')
      expect(result).toEqual([])
    })

    test('endKey greater than now', () => {
      const result = util.keysInRange('2018', '2025')
      expect(result).toEqual(['2018', '2019', '2020', '2021', '2022', '2023', '2024', '2025'])
    })

    test('quarterly endKey greater than now', () => {
      MockDate.set('2023-07-10')
      const result = util.keysInRange('2023Q1', '2024Q1')
      expect(result).toEqual(['2023Q1', '2023Q2', '2023Q3', '2023Q4', '2024Q1'])
    })

    test('throw if key mismatch', () => {
      expect(() => {
        util.keysInRange('2001', '2001Q2')
      }).toThrow()
    })
  })

  describe('getKeysByScope()', () => {
    beforeEach(() => {
      util.keysByScope.annually = [] // clear cache
      util.keysByScope.quarterly = [] // clear cache
      util.keysByScope.monthly = [] // clear cache
      util.keysByScope.daily = [] // clear cache
    })
    afterEach(() => {
      MockDate.reset()
    })
    test('generate annually scope by key', () => {
      MockDate.set('2004-05-22')
      const keys = util.getKeysByScope('annually')
      expect(keys).toEqual(['2000', '2001', '2002', '2003', '2004', '2005', '2006'])
    })
    test('generate quarterly scope by key', () => {
      MockDate.set('2002-11-22')
      const keys = util.getKeysByScope('quarterly')
      expect(keys).toEqual([
        '2000Q1',
        '2000Q2',
        '2000Q3',
        '2000Q4',
        '2001Q1',
        '2001Q2',
        '2001Q3',
        '2001Q4',
        '2002Q1',
        '2002Q2',
        '2002Q3',
        '2002Q4',
        '2003Q1',
        '2003Q2'
      ])
    })
    test('generate monthly scope by key', () => {
      MockDate.set('2000-05-22')
      const keys = util.getKeysByScope('monthly')
      expect(keys).toEqual(['2000-01', '2000-02', '2000-03', '2000-04', '2000-05', '2000-06', '2000-07'])
    })
    test('generate daily scope by key', () => {
      MockDate.set('2000-01-10')
      const keys = util.getKeysByScope('daily')
      expect(keys).toEqual([
        '2000-01-01',
        '2000-01-02',
        '2000-01-03',
        '2000-01-04',
        '2000-01-05',
        '2000-01-06',
        '2000-01-07',
        '2000-01-08',
        '2000-01-09',
        '2000-01-10',
        '2000-01-11',
        '2000-01-12'
      ])
    })
    test('get scope from cache', () =>{
      MockDate.set('2004-05-22')
      jest.spyOn(util, 'formatScope')
      const keys = util.getKeysByScope('annually')
      expect(keys).toEqual(['2000', '2001', '2002', '2003', '2004', '2005', '2006'])
      expect(util.formatScope).toHaveBeenCalled()

      util.formatScope.mockReset()
      const keys2 = util.getKeysByScope('annually')
      expect(keys2).toEqual(['2000', '2001', '2002', '2003', '2004', '2005', '2006'])
      expect(util.formatScope).not.toHaveBeenCalled()
      util.formatScope.mockRestore()
    })
  })

  describe('forwardFill()', () => {
    test('normal case', () => {
      const rawResult = util.forwardFill([
        { seen: new Date('2016-02-10'), v: 1 },
        { seen: new Date('2016-02-15'), v: 2 }
      ])

      const result = rawResult.map(value => ({
        v: value.v,
        seen: value.seen
      }))

      expect(result).toEqual([
        { seen: new Date('2016-02-10'), v: 1 },
        { seen: new Date('2016-02-11'), v: 1 },
        { seen: new Date('2016-02-12'), v: 1 },
        { seen: new Date('2016-02-13'), v: 1 },
        { seen: new Date('2016-02-14'), v: 1 },
        { seen: new Date('2016-02-15'), v: 2 }
      ])
    })
    test('minimum seen use last value before minSeen', () => {
      const options = { minSeen: new Date('2016-02-12') }
      const rawResult = util.forwardFill(
        [
          { seen: new Date('2016-02-10'), v: 1 },
          { seen: new Date('2016-02-15'), v: 2 }
        ],
        options
      )

      const result = rawResult.map(value => ({
        v: value.v,
        seen: value.seen
      }))

      expect(result).toEqual([
        { seen: new Date('2016-02-12'), v: 1 },
        { seen: new Date('2016-02-13'), v: 1 },
        { seen: new Date('2016-02-14'), v: 1 },
        { seen: new Date('2016-02-15'), v: 2 }
      ])
    })

    test('options.fillToDate', () => {
      const options = {
        minSeen: new Date('2016-02-12'),
        fillToDate: new Date('2016-02-17')
      }

      const rawResult = util.forwardFill(
        [
          { seen: new Date('2016-02-10'), v: 1 },
          { seen: new Date('2016-02-15'), v: 2 }
        ],
        options
      )

      const result = rawResult.map(value => ({
        v: value.v,
        seen: value.seen
      }))

      expect(result).toEqual([
        { seen: new Date('2016-02-12'), v: 1 },
        { seen: new Date('2016-02-13'), v: 1 },
        { seen: new Date('2016-02-14'), v: 1 },
        { seen: new Date('2016-02-15'), v: 2 },
        { seen: new Date('2016-02-16'), v: 2 },
        { seen: new Date('2016-02-17'), v: 2 }
      ])
    })
  })

  describe('convertkey()', () => {
    test('convert null key', () => {
      const result = util.convertKey(null, 'quarterly')
      expect(result).toEqual(null)
    })

    describe('from annually', () => {
      test('convert annually to annually', () => {
        const result = util.convertKey('2018', 'annually')
        expect(result).toEqual('2018')
      })
      test('convert annually to quarterly', () => {
        const result = util.convertKey('2018', 'quarterly')
        expect(result).toEqual('2018Q4')
      })
      test('convert annually to monthly', () => {
        const result = util.convertKey('2018', 'monthly')
        expect(result).toEqual('2018-12')
      })
      test('convert annually to daily', () => {
        const result = util.convertKey('2018', 'daily')
        expect(result).toEqual('2018-12-31')
      })
    })

    describe('from quarterly', () => {
      test('convert quarterly to annually', () => {
        const result = util.convertKey('2018Q1', 'annually')
        expect(result).toEqual('2018')
      })
      test('convert quarterly to quarterly', () => {
        const result = util.convertKey('2018Q1', 'quarterly')
        expect(result).toEqual('2018Q1')
      })
      test('convert quarterly to monthly', () => {
        const result = util.convertKey('2018Q2', 'monthly')
        expect(result).toEqual('2018-6')
      })
      test('convert quarterly to daily', () => {
        const result = util.convertKey('2018Q2', 'daily')
        expect(result).toEqual('2018-6-30')
      })
    })

    describe('from monthly', () => {
      test('convert monthly to annually', () => {
        const result = util.convertKey('2018-3', 'annually')
        expect(result).toEqual('2018')
      })
      test('convert monthly to quarterly', () => {
        const result = util.convertKey('2018-3', 'quarterly')
        expect(result).toEqual('2018Q1')
      })
      test('convert monthly to monthly', () => {
        const result = util.convertKey('2018-3', 'monthly')
        expect(result).toEqual('2018-3')
      })
      test('convert monthly to daily', () => {
        const result = util.convertKey('2018-3', 'daily')
        expect(result).toEqual('2018-3-31')
      })
    })

    describe('from daily', () => {
      test('convert daily to annually', () => {
        const result = util.convertKey('2018-3-15', 'annually')
        expect(result).toEqual('2018')
      })
      test('convert daily to quarterly', () => {
        const result = util.convertKey('2018-3-15', 'quarterly')
        expect(result).toEqual('2018Q1')
      })
      test('convert daily to monthly', () => {
        const result = util.convertKey('2018-3-15', 'monthly')
        expect(result).toEqual('2018-3')
      })
      test('convert daily to daily', () => {
        const result = util.convertKey('2018-3-15', 'daily')
        expect(result).toEqual('2018-3-15')
      })
    })
  })

  describe('forwardFillObject()', () => {
    test('default', () => {
      const toFill = {
        '2017Q1': 1,
        '2017Q2': 2
      }
      const currentDate = new Date('2017-09-30')
      const filledObj = util.forwardFillObject(toFill, currentDate)
      expect(filledObj).toEqual({
        '2017Q1': 1,
        '2017Q2': 2,
        '2017Q3': 2
      })
    })

    describe('options.maxLookbackPeriod', () => {
      test('no limit', () => {
        const toFill = {
          '2017Q1': 1,
          '2017Q2': 2
        }
        const currentDate = new Date('2018-01-01')
        const filledObj = util.forwardFillObject(toFill, currentDate)
        expect(filledObj).toEqual({
          '2017Q1': 1,
          '2017Q2': 2,
          '2017Q3': 2,
          '2017Q4': 2
        })
      })

      test('greater than limit', () => {
        const toFill = {
          '2017Q1': 1,
          '2017Q2': 2
        }
        const currentDate = new Date('2018-01-01')
        const filledObj = util.forwardFillObject(toFill, currentDate, { maxLookbackPeriod: 1 })
        expect(filledObj).toEqual({
          '2017Q1': 1,
          '2017Q2': 2
        })
      })

      test('equal to limit', () => {
        const toFill = {
          '2017Q1': 1,
          '2017Q2': 2
        }
        const currentDate = new Date('2018-01-01')
        const filledObj = util.forwardFillObject(toFill, currentDate, { maxLookbackPeriod: 2 })
        expect(filledObj).toEqual({
          '2017Q1': 1,
          '2017Q2': 2,
          '2017Q3': 2,
          '2017Q4': 2
        })
      })
    })
  })

  describe('differenceMonth', () => {
    test('same date', () => {
      const date = new Date()
      const result = util.differenceMonth(date, date)
      expect(result).toBe(0)
    })
    test('date in same year', () => {
      const a = new Date('2017-01-01')
      const b = new Date('2017-02-01')
      const result = util.differenceMonth(b, a)
      expect(result).toBe(1)
    })

    test('date with different year', () => {
      const a = new Date('2017-01-01')
      const b = new Date('2018-02-01')
      const result = util.differenceMonth(b, a)
      expect(result).toBe(13)
    })

    test('first date smaller', () => {
      const a = new Date('2017-01-01')
      const b = new Date('2018-02-01')
      const result = util.differenceMonth(a, b)
      expect(result).toBe(-13)
    })
  })

  describe('generateDates', () => {
    test('result in end of period dates', () => {
      const scope = 'quarterly'
      const a = new Date('2017-01-01')
      const b = new Date('2018-02-01')
      const result = util.generateDates(a, b, scope)
      expect(result).toEqual([
        new Date('2017-03-31'),
        new Date('2017-06-30'),
        new Date('2017-09-30'),
        new Date('2017-12-31')
      ])
    })
    test('result in end of period dates include to date', () => {
      const scope = 'monthly'
      const a = new Date('2017-11-01')
      const b = new Date('2018-03-31')
      const result = util.generateDates(a, b, scope)
      expect(result).toEqual([
        new Date('2017-11-30'),
        new Date('2017-12-31'),
        new Date('2018-01-31'),
        new Date('2018-02-28'),
        new Date('2018-03-31')
      ])
    })
  })
})
