// eslint-disable-next-line prettier/prettier
const indexOf = [].indexOf || function _indexOf(item) { for (let i = 0, l = this.length; i < l; i += 1) { if (i in this && this[i] === item) return i; } return -1; }

const _last = require('lodash/last')
const _get = require('lodash/get')

const dateFns = require('date-fns')
// const subMinutes = require('date-fns/subMinutes')
// // const format = require('date-fns/lightFormat')
// const format = require('date-fns/format')
// const lightFormat = require('date-fns/lightFormat')
// const parse = require('date-fns/parse')
// const startOfDay = require('date-fns/startOfDay')

const { subMinutes, format, lightFormat, parse, startOfDay } = dateFns

const debug = require('debug')('tfdata:util')

const MONTH_RE = /^(\d{4})-(1[012]|0?[1-9])$/g
const QUARTER_RE = /^(\d{4})[Q]([1-4]{1})$/g
const FISCAL_RE = /^(\d{4})$/g
const DAY_RE = /^(\d{4})-(\d{1}\d{0,1})-(\d{1}\d{0,1})$/g
const SCOPE_TO_SUFFIX = {
  annually: 'year',
  quarterly: 'quarter',
  monthly: 'month',
  daily: 'day'
}

// scopeKey from Source (project-k)
const SCOPE_TO_STRING_FORMAT = {
  annually: 'yyyy',
  quarterly: 'yyyyQQQ',
  monthly: 'yyyy-M',
  daily: 'yyyy-M-d'
}

// normalized Scopekey
const SCOPE_TO_KEY_FORMAT = {
  annually: 'yyyy',
  quarterly: 'yyyyQQQ',
  monthly: 'yyyy-MM',
  daily: 'yyyy-MM-dd'
}

const AVAILABLE_SCOPES = ['annually', 'quarterly', 'monthly', 'daily', 'timeless']

const START_YEAR = '2000'
const ADDED_FUTURE_PERIOD = 2

const util = {
  keysByScope: {
    annually: [],
    quarterly: [],
    monthly: [],
    daily: []
  },
  endOf: {
    year: dateFns.endOfYear,
    quarter: dateFns.endOfQuarter,
    month: dateFns.endOfMonth,
    day: dateFns.endOfDay
  },
  toUTC(date) {
    return subMinutes(date, date.getTimezoneOffset())
  },
  /**
   * Set value
   * @param {Object} obj      Object to set value of
   * @param {String} scopeKey Key to set
   * @param {Any} val      Value of the Key
   * @param {Object} options  =             {} [description]
   */
  DATEKEY_FORMAT: 'YYYY-M-D',

  _getValue(obj, path, scopeKey, condition) {
    let cache
    let target
    if (path != null) {
      cache = _get(obj, path) || null
    } else {
      cache = obj
    }
    if (cache == null) {
      return null
    }
    if (cache._type != null) {
      if (cache._type === 'timeless') {
        return cache.value
      }
      target = cache != null ? cache._values.find(condition) : undefined
      if (target != null) {
        return target.v
      }
      return null
    }
    return cache
  },

  getFiscalValue(obj, path, fiscalKey) {
    return util._getValue(obj, path, fiscalKey, val => {
      return val.fk === fiscalKey
    })
  },

  getCalendarValue(obj, path, calendarKey) {
    let cache
    let values
    const date = util.scopeKeyToDate(calendarKey)
    if (path != null) {
      cache = _get(obj, path) || null
    } else {
      cache = obj
    }
    if (cache == null) {
      return null
    }
    if (cache._type != null) {
      if (cache._type === 'timeless') {
        return cache.value
      }
      values = cache._values.filter(v => {
        return v.seen <= date
      })
      const last = _last(values)
      return last != null ? last.v : undefined
    }
    return cache
  },

  addDay(date, i) {
    const result = new Date(date)
    result.setDate(date.getDate() + i)
    return result
  },
  addMonth(date, i) {
    return dateFns.addMonths(date, i)
  },
  addQuarter(date, i) {
    return dateFns.addQuarters(date, i)
  },
  addYear(date, i) {
    const result = new Date(date)
    result.setYear(date.getFullYear() + i)
    return result
  },
  subtractDay(date, i) {
    return util.addDay(date, -i)
  },
  subtractMonth(date, i) {
    return util.addMonth(date, -i)
  },
  subtractQuarter(date, i) {
    return util.addQuarter(date, -i)
  },
  subtractYear(date, i) {
    return util.addYear(date, -i)
  },
  addDuration(date, i, scope) {
    switch (scope) {
      case 'annually':
        return util.addYear(date, i)
      case 'quarterly':
        return util.addQuarter(date, i)
      case 'monthly':
        return util.addMonth(date, i)
      case 'daily':
        return util.addDay(date, i)
      default:
        throw new TypeError(`Unsupported Scope. ${scope}`)
    }
  },
  getQuarter(date) {
    return Math.ceil((date.getMonth() + 1) / 3)
  },
  subtractDuration(date, i, scope) {
    return util.addDuration(date, -i, scope)
  },

  formatScope(scope, date, normalize = false) {
    const formatString = normalize ? SCOPE_TO_KEY_FORMAT[scope] : SCOPE_TO_STRING_FORMAT[scope]
    if (scope === 'quarterly') {
      return format(date, formatString)
    }
    // lighformat is faster but has no Quarter format
    return lightFormat(date, formatString)
  },

  convertKey(key, toScope, _fromScope) {
    if (key == null) {
      return null
    }
    const fromScope = _fromScope || util.scopeKeyType(key)
    if (fromScope === toScope) return key
    const fromFormat = SCOPE_TO_STRING_FORMAT[fromScope]
    const fromPeriod = SCOPE_TO_SUFFIX[fromScope]
    const fromDate = util.toUTC(parse(key, fromFormat, new Date()))
    return util.formatScope(toScope, startOfDay(util.endOf[fromPeriod](fromDate)))
  },

  // deprecated
  scopeToMomentTimeKey(scope) {
    if (indexOf.call(AVAILABLE_SCOPES, scope) < 0) {
      throw new TypeError(`[Util] Illegal formula scope(${JSON.stringify(scope)})`)
    }
    if (scope === 'timeless') {
      throw new TypeError(`[Util] Illegal formula scope(${scope})`)
    }
    return SCOPE_TO_SUFFIX[scope]
  },

  scopeToTimeKey(scope) {
    if (indexOf.call(AVAILABLE_SCOPES, scope) < 0) {
      throw new TypeError(`[Util] Illegal formula scope(${JSON.stringify(scope)})`)
    }
    if (scope === 'timeless') {
      throw new TypeError(`[Util] Illegal formula scope(${scope})`)
    }
    return SCOPE_TO_SUFFIX[scope]
  },

  stringFormatForScope(scope) {
    if (indexOf.call(AVAILABLE_SCOPES, scope) < 0) {
      throw new TypeError(`[Util] Illegal formula scope(${scope})`)
    }
    if (scope === 'timeless') {
      throw new TypeError(`[Util] Illegal formula scope(${scope})`)
    }
    return SCOPE_TO_STRING_FORMAT[scope]
  },

  /**
   * Convert a date to a key period that this date will be first
   * seen. e.g. 2016-07-20 => '2016Q3'
   * @param  {Date} date
   * @param  {String} scopeType Type of key to output
   * @return {String}           key
   */

  dateToScopeKey(date, scopeType) {
    if (date == null) return null
    if (scopeType === 'timeless') {
      return null
    }
    const timeKey = util.scopeToTimeKey(scopeType)
    return util.formatScope(scopeType, startOfDay(util.endOf[timeKey](util.toUTC(date))))
  },

  dateToNormalizeScopeKey(date, scopeType) {
    if (scopeType === 'timeless') {
      return null
    }
    const timeKey = util.scopeToTimeKey(scopeType)
    return util.formatScope(scopeType, startOfDay(util.endOf[timeKey](util.toUTC(date))), true)
  },

  normalizeKey(key) {
    if (key.match(MONTH_RE)) {
      if (key.length === 6) {
        return key.replace('-', '-0')
      }
    }
    return key
  },

  /**
   * Convert key to scopeType
   * @param  {[type]} key [description]
   * @return {[type]}     [description]
   */
  scopeKeyType(_key) {
    if (_key != null) {
      const key = _key.toString()
      if (key.match(FISCAL_RE) != null) {
        return 'annually'
      }
      if (key.match(QUARTER_RE) != null) {
        return 'quarterly'
      }
      if (key.match(MONTH_RE) != null) {
        return 'monthly'
      }
      if (key.match(DAY_RE) != null) {
        return 'daily'
      }
      return null
    }
    return 'timeless'
  },

  scopeKeyToDate(_key, scope) {
    if (_key == null) {
      return null
    }
    const key = _key.toString()
    const scopeType = scope || util.scopeKeyType(key)
    const fromFormat = SCOPE_TO_STRING_FORMAT[scopeType]
    const period = util.scopeToTimeKey(scopeType)
    return util.toUTC(startOfDay(util.endOf[period](parse(key, fromFormat, new Date()))))
  },

  addScopeKey(key, i = 1) {
    const type = util.scopeKeyType(key)
    if (type === 'timeless') {
      return null
    }
    const date = util.scopeKeyToDate(key)
    const addedDate = util.addDuration(date, i, type)
    return util.dateToScopeKey(addedDate, type)
  },

  addNormalizedScopeKey(key, i = 1, scope) {
    const type = scope || util.scopeKeyType(key)
    if (type === 'timeless') {
      return null
    }
    const date = util.scopeKeyToDate(key, type)
    const addedDate = util.addDuration(date, i, type)
    return util.dateToNormalizeScopeKey(addedDate, type)
  },

  subtractScopeKey(key, _i = 1) {
    let i = _i
    if (i == null) {
      i = 1
    }
    return util.addScopeKey(key, -i)
  },

  scopeConversionTable: {
    monthly: {
      monthly: 1,
      quarterly: 1 / 3,
      annually: 1 / 12
    },
    quarterly: {
      monthly: 3,
      quarterly: 1,
      annually: 1 / 4
    },
    annually: {
      monthly: 12,
      quarterly: 4,
      annually: 1
    }
  },

  /**
   * Convert from one scope period unit to another
   * e.g. 1 annually -> 4 quarterly -> 12 monthly
   * @param  {Integer} i                 [description]
   * @param  {String} fromType [description]
   * @param  {String} toType   [description]
   * @return {Integer}                   [description]
   */
  convertScopePeriod(i, fromType, toType) {
    util.log(`convert type ${fromType} to type ${toType} for period ${i}`)
    if (fromType === toType) {
      return i
    }
    const mod = util.scopeConversionTable[fromType][toType]
    return Math.ceil(i * mod)
  },

  compareKey(a, b) {
    if (b == null) {
      return 1
    }
    if (a == null) {
      return -1
    }
    if (a === b) {
      return 0
    }
    if (util.scopeKeyType(a) !== 'monthly') {
      if (a > b) {
        return 1
      }
      return -1
      // monthly compare
    }

    const [aYear, aMonth] = a.split('-')
    const [bYear, bMonth] = b.split('-')
    if (aYear === bYear) {
      if (aMonth > bMonth) {
        return 1
      }
      return -1
    }
    if (aYear > bYear) {
      return 1
    }
    return -1
  },

  log(msg) {
    if (typeof window !== 'undefined' && window !== null) {
      if (window.debug != null) {
        if (window.debug) {
          return debug(msg)
        }
      } else {
        return debug(msg)
      }
    } else if (debug != null) {
      return debug(msg)
    }
    return false
  },

  sortSeen(a, b) {
    return a.seen - b.seen
  },

  escapeMongoPath(path) {
    return path.replace(/\./g, '\uff0e')
  },

  generateDates(from, to, scopeType) {
    // console.log('fromDate' , from)
    const periodType = util.scopeToTimeKey(scopeType)
    let currentDate = startOfDay(util.endOf[periodType](from))
    const dates = []
    while (currentDate <= to) {
      dates.push(util.toUTC(currentDate))
      currentDate = startOfDay(util.endOf[periodType](util.addDuration(currentDate, 1, scopeType)))
    }
    return dates
  },

  // generate from past to now
  getKeysByScope(scope) {
    if (this.keysByScope[scope].length === 0) {
      const startDate = new Date(START_YEAR)
      const keys = []
      let currentDate = new Date(startDate)
      // const endDate = new Date()
      const endDate = util.addDuration(new Date(), ADDED_FUTURE_PERIOD, scope)
      // console.log(newEndDate)
      while (endDate >= currentDate) {
        keys.push(util.formatScope(scope, currentDate, true))
        currentDate = util.addDuration(currentDate, 1, scope)
      }
      this.keysByScope[scope] = keys
    }
    return this.keysByScope[scope]
  },

  // potential bottleneck function (probably impossible to optimize)
  // fix by caching keys
  keysInRange(startKey, endKey) {
    const safeStartKey = util.normalizeKey(startKey)
    const safeEndKey = util.normalizeKey(endKey)
    const scope = util.scopeKeyType(safeStartKey)
    // const scope = 'quarterly'
    if (!util.isSameKeyType(safeStartKey, safeEndKey)) {
      throw new Error(`Mismatch key type: ${startKey} and ${endKey}`)
    }

    if (safeStartKey === safeEndKey) {
      return [safeStartKey]
    }
    if (safeStartKey >= safeEndKey) {
      return []
    }

    let keys = this.getKeysByScope(scope)
    keys = keys.filter(key => key >= safeStartKey && key <= safeEndKey)
    return keys
  },

  isSameKeyType(a, b) {
    const aType = util.scopeKeyType(a)
    const bType = util.scopeKeyType(b)
    return aType === bType
  },

  /**
   *
   * @param {*} obj
   * @param {*} endDate
   * @param {*} options
   * @param {*} options.maxLookbackPeriod
   */
  forwardFillObject(obj, endDate, options = {}) {
    const filledObject = { ...obj }
    let lastKey = _last(Object.keys(filledObject).sort())
    if (lastKey == null) {
      return obj
    }
    const lastValue = obj[lastKey]

    const scopeType = util.scopeKeyType(lastKey)
    let endKey = util.dateToNormalizeScopeKey(endDate, scopeType)

    // back 1 period when endDate is not the end of period
    if (endDate < util.scopeKeyToDate(endKey)) {
      endKey = util.addNormalizedScopeKey(endKey, -1)
    }
    lastKey = util.addNormalizedScopeKey(lastKey, 1)
    const keysToFill = util.keysInRange(lastKey, endKey)
    if (options.maxLookbackPeriod != null) {
      const minLastKey = keysToFill.at(-options.maxLookbackPeriod)
      const lastKeyLessThanMin = util.compareKey(minLastKey, lastKey) === 1
      if (lastKeyLessThanMin) {
        return filledObject
      }
    }

    keysToFill.forEach(key => {
      filledObject[key] = lastValue
    })

    return filledObject
  },

  // not in used
  // no optimized
  forwardFill(
    array,
    options = {
      minSeen: new Date('2002-01-01'),
      fillToDate: null
    }
  ) {
    const filledArray = []
    const map = {}
    const seens = []

    array.forEach(v => {
      map[v.seen] = v
      seens.push(v.seen)
    })

    seens.sort((a, b) => a - b)

    let firstDate = null
    if (seens[0] < options.minSeen) {
      firstDate = options.minSeen
      map[firstDate] = map[seens[0]]
    } else {
      ;[firstDate] = seens
    }

    let { fillToDate } = options
    const lastDate = seens[seens.length - 1]
    if (fillToDate == null) {
      fillToDate = lastDate
    }

    const currentDate = new Date(firstDate)
    let lastSeenData = map[firstDate]

    while (currentDate <= fillToDate) {
      let data = map[currentDate]
      if (data == null) {
        data = { ...lastSeenData }
        data._t = new Date(lastSeenData.seen)
        data.seen = new Date(currentDate)
      } else {
        lastSeenData = data
      }

      const currentData = {
        v: data.v,
        _t: data._t,
        seen: new Date(currentDate)
      }

      filledArray.push(currentData)
      currentDate.setDate(currentDate.getDate() + 1)
    }

    return filledArray
  },

  /**
   *
   * @param {Date} a - the newer date
   * @param {Date} b - the older date
   * @return {Number} the difference in month of two date
   */
  differenceMonth(a, b) {
    const yearOnlyDiff = a.getFullYear() - b.getFullYear()
    const monthOnlyDiff = a.getMonth() - b.getMonth()
    const monthDiff = yearOnlyDiff * 12 + monthOnlyDiff
    return monthDiff
  }
}

module.exports = util
